using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;

namespace klient
{
	/// <summary>
	/// Pripojeni k severu.
	/// </summary>
	class Klient
	{
		/// <summary>
		/// Je pripojeno?
		/// </summary>
		private bool pripojeno;

		/// <summary>
		/// Uchovava jmeno serveru.
		/// </summary>
		private string server;

		/// <summary>
		/// Uchovava cislo portu.
		/// </summary>
		private int port;

		/// <summary>
		/// Uchovava TCP clienta
		/// </summary>
		private TcpClient client;

		/// <summary>
		/// Cteni ze socketu
		/// </summary>
		private StreamReader sr;

		/// <summary>
		/// Zapis na socket  - StreamWriter zahadne nefunguje :(.
		/// </summary>
		private Stream stream;

		/// <summary>
		/// Vsechny vytahy v systemu.
		/// </summary>
		private ArrayList vytahy;

		/// <summary>
		/// Text box pro vystup dat ze serveru.
		/// </summary>
		private System.Windows.Forms.ListBox logs;

		/// <summary>
		/// Vytvori pripojeni k serveru.
		/// </summary>
		public Klient(System.Windows.Forms.ListBox logs)
		{
			this.logs = logs;
			this.pripojeno = false;
			this.vytahy = new ArrayList();
		}

		/// <summary>
		/// Pripoji se k serveru.
		/// </summary>
		/// <param name="server">Jmeno serveru.</param>
		/// <param name="port">Port pripojeni.</param>
		public void Pripoj(string server, int port)
		{
			// je navazano spojeni?
			if (this.pripojeno)
				return;
			
			// zapamatuje, vytvori klienta a vytvori si stream writer
			// a stream reader pro snadnejsi ovladani serveru
			this.server = server;
			this.port = port;
			this.client = new TcpClient(server,port);
			this.stream = client.GetStream();
			this.sr = new StreamReader(stream);

			this.pripojeno = true;

			// spusti ovladani serveru jako nove vlakno
			new Thread(new ThreadStart(this.OvladejServer)).Start();
			//this.OvladejServer();
		}


		/// <summary>
		/// Odpoji se od serveru.
		/// </summary>
		public void Odpoj()
		{
			if (!this.pripojeno)
				return;
			this.client.Close();
		}

		/// <summary>
		/// Odesle prikaz serveru.
		/// </summary>
		/// <param name="prikaz">Prikaz pro server.</param>
		public void OdesliPrikaz(string prikaz)
		{
			if (!this.pripojeno)
				return;
			
			byte[] b = Encoding.ASCII.GetBytes(prikaz);
			this.stream.Write(b,0,b.Length);
			this.stream.WriteByte(13);
			this.stream.WriteByte(10);

			Zapis("Poslano: " + prikaz);
		}

		/// <summary>
		/// Precte odpoved na prikaz pro server.
		/// </summary>
		/// <returns>Odpoved serveru.</returns>
		private string PrectiOdpoved()
		{
			string s = sr.ReadLine();
			Zapis("Obdrzeno: " + s);
			return s;
		}

		/// <summary>
		/// Zapise do listboxu s daty vystup ze serveru.
		/// </summary>
		/// <param name="text">Co zapsat.</param>
		private void Zapis(string text)
		{
			this.logs.Items.Add(text);
		}

		/// <summary>
		/// Ridi server odeslianim prikazu.
		/// </summary>
		private void OvladejServer()
		{
			// chyta chyby
			try
			{
				// inicializuje a vytvori patra

				this.OdesliPrikaz("shafts");
				int pocetVytahu = Convert.ToInt32(this.PrectiOdpoved().Split(new char[] {' '})[1]);
				this.OdesliPrikaz("levels");
				int pocetPater = Convert.ToInt32(this.PrectiOdpoved().Split(new char[] {' '})[1]);

				for (int a = 0; a < pocetVytahu; a++)
				{
					this.vytahy.Add(new Vytah(pocetPater,a));
					this.OtevrenVytah += new otevrenVytah(((Vytah)this.vytahy[a]).Otevren);
					this.OtevreniDveri += new otevreniDveri(((Vytah)this.vytahy[a]).OtevriDvere);
					this.PresunKabiny += new presunKabiny(((Vytah)this.vytahy[a]).Presun);
				}

				// ovlada server
				string odpoved;
				string[] odpovedCasti;

				while (true)
				{
					odpoved = this.PrectiOdpoved();
					odpovedCasti = odpoved.Split(new char[] {' '});

					// jednotlive udalosti
					switch (odpovedCasti[0])
					{
						case "levelbuttonup":
							OdesliPrikaz("indiupon " + odpovedCasti[1]);
							this.NajdiVytah(Convert.ToInt32(odpovedCasti[1]));
							break;
						case "levelbuttondn":
							OdesliPrikaz("indidnon " + odpovedCasti[1]);
							this.NajdiVytah(Convert.ToInt32(odpovedCasti[1]));
							break;
						case "cabinbutton":
							this.PresunKabiny(this, new VytahEventArgs(Convert.ToInt32(odpovedCasti[1]),Convert.ToInt32(odpovedCasti[2])));
							break;
						case "opened":
							this.OtevrenVytah(this, new VytahEventArgs(Convert.ToInt32(odpovedCasti[1]), 0));
							break;
						default:
							//Zapis("Prikaz zatim neimplementovan!");
							break;
					}
				}
			}
			catch (Exception e)
			{
				System.Windows.Forms.MessageBox.Show("Chyba pi zpracovn pkaz!\n\n"+e.Message);
			}
		}

		/// <summary>
		/// Najde prvni volny vytah.
		/// </summary>
		/// <param name="patro">Vyhledavane patro.</param>
		private void NajdiVytah(int patro)
		{
			// hleda vytah na patre
			foreach (Vytah v in this.vytahy)
			{
				if (v.JeNaPatre(patro))
				{
					this.OtevreniDveri(this,new VytahEventArgs(v.CisloVytahu,0));
					return;
				}
			}
			// jinak hleda nepracujici vytah
			foreach (Vytah v in this.vytahy)
			{
				if (v.NemaPraci)
				{
					this.PresunKabiny(this,new VytahEventArgs(v.CisloVytahu,patro));
					this.OtevreniDveri(this,new VytahEventArgs(v.CisloVytahu,0));
				}
			}
			// jinak ma cestujuci smulu a bude si muset pockat :o)
		}

		#region Zastupci a udalosti

		/// <summary>
		/// Udalost od serveru - otevreny dvere
		/// </summary>
		public delegate void otevrenVytah(object sender, VytahEventArgs e);
		/// <summary>
		/// Event handler pro udalost otevreni dveri (od serveru).
		/// </summary>
		public event otevrenVytah OtevrenVytah;

		/// <summary>
		/// Udalost od klienta - je nutne otevrit dvere
		/// </summary>
		public delegate void otevreniDveri(object sender, VytahEventArgs e);
		/// <summary>
		/// Event handler pro otevreni dveri na prikaz klienta.
		/// </summary>
		public event otevreniDveri OtevreniDveri;

		/// <summary>
		/// Presune kabinu na prikaz klienta.
		/// </summary>
		public delegate void presunKabiny(object sender, VytahEventArgs e);

		/// <summary>
		/// Event handler pro presun kabiny na prikaz klienta.
		/// </summary>
		public event presunKabiny PresunKabiny;

		#endregion
	}

	#region Trida VytahEventArgs urcena k predavani udalosti

	/// <summary>
	/// Trida dulezita k zachyceni udalosti.
	/// </summary>
	public class VytahEventArgs : EventArgs
	{
		/// <summary>
		/// Cislo vytahu.
		/// </summary>
		public int vytah;

		/// <summary>
		/// Cislo patra.
		/// </summary>
		public int patro;

		/// <summary>
		/// Vytvori nove parametry udalosti.
		/// </summary>
		/// <param name="vytah"></param>
		/// <param name="patro"></param>
		public VytahEventArgs(int vytah, int patro)
		{
			this.vytah = vytah;
			this.patro = patro;
		}
	}

	#endregion

	#region Trida Vytah

	/// <summary>
	/// Vytah.
	/// </summary>
	class Vytah
	{
		/// <summary>
		/// Pocet poschodi, do kterych vytah jezdi.
		/// </summary>
		private int pocetPoschodi;

		/// <summary>
		/// Cislo vytahu.
		/// Pozdeji se pouziva ve specialnich prikazech pro konkretni vytah
		/// </summary>
		private int cisloVytahu;

		/// <summary>
		/// Cislo vytahu.
		/// </summary>
		public int CisloVytahu
		{
			get { return this.cisloVytahu; }
		}

		/// <summary>
		/// V jakem patre je vytah umisten;
		/// </summary>
		private int aktualniPatro;

		/// <summary>
		/// Zapamatuje si vsechny stisknuta tlacitka a snazi se udalosti zbavit co nejrychleji
		/// </summary>
		private ArrayList zasobnikPresunu;

		/// <summary>
		/// Pokud nema praci, vraci true
		/// </summary>
		public bool NemaPraci
		{
			get { return ( this.zasobnikPresunu.Count == 0 ? true : false); }
		}

		/// <summary>
		/// Vytvori novy vytah.
		/// </summary>
		/// <param name="pocetPoschodi">Pocet poschodi vytahu</param>
		/// <param name="cisloVytahu">Cislo vytahu</param>
		public Vytah(int pocetPoschodi, int cisloVytahu)
		{
			this.pocetPoschodi = pocetPoschodi;
			this.cisloVytahu = cisloVytahu;
			this.aktualniPatro = 0;
			this.zasobnikPresunu = new ArrayList();
		}

		/// <summary>
		/// Je vytah na patre?
		/// </summary>
		/// <param name="patro">Cislo patra</param>
		/// <returns>True ANO jinak false</returns>
		public bool JeNaPatre(int patro)
		{
			if (this.aktualniPatro == patro)
				return true;
			else
				return false;
		}

		/// <summary>
		/// Udalost otevreni vytahu
		/// </summary>
		/// <param name="sender">Udalost vyslana</param>
		/// <param name="e">Parametry</param>
		public void Otevren(object sender, VytahEventArgs e)
		{
			// pouze tento vytah
			if (e.vytah != this.cisloVytahu)
				return;

			// vypne svititka
			((Klient)sender).OdesliPrikaz("indiupoff " + e.patro.ToString());
			((Klient)sender).OdesliPrikaz("indidnoff " + e.patro.ToString());

			// na tomto patre vyklada cestujici
			// jednoduzsi zpusob jsem nenasel :o(
			for (int a = 0; a <= this.zasobnikPresunu.Count ; a++)
			{
				this.zasobnikPresunu.Remove(e.patro);
			}
		}

		/// <summary>
		/// Udalost presun kabiny - tedi stisk tlacitka v kabine
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		public void Presun(object sender, VytahEventArgs e)
		{
			// pouze tento vytah
			if (e.vytah != this.cisloVytahu)
				return;

			this.zasobnikPresunu.Add(e.patro);

			this.aktualniPatro = e.patro;
			((Klient)sender).OdesliPrikaz("floor " + this.cisloVytahu + " " + e.patro);
		}

		/// <summary>
		/// Otevre dvere na zavolani klienta.
		/// </summary>
		/// <param name="sender">Odesilatel</param>
		/// <param name="e">Parametry</param>
		public void OtevriDvere(object sender, VytahEventArgs e)
		{
			// pouze tento vytah
			if (e.vytah != this.cisloVytahu)
				return;

			((Klient)sender).OdesliPrikaz("open " + this.cisloVytahu);
		}
	}
	#endregion
}
